;;; -*- mode: lisp; syntax: common-lisp; base: 10; package: JSGEN -*-
;;
;; expressions.lisp -- Macros for defining functions that print s-expressions
;;                     to a C-like language.
;;
;; Copyright (c) 2006, Jason Kantz <jason@kantz.com>
;; All rights reserved.
;;
;;; Code:

(in-package :jsgen)


;;; expressions (op . args)

(defun primitivep (x)
  (atom x))

(defun expressionp (x)
  (consp x))

(defun operator (expression)
  (car expression))

(defun arguments (expression)
  (cdr expression))




;;; operator properties

(defun operator-type (op)
  (get op '%operator-type))

(defun (setf operator-type) (type op)
  (assert (member type '(:infix-operator :special-statement :special-expression :macro-operator)))
  (setf (get op '%operator-type) type))

(defun pretty-function (op)
  (get op '%operator-pretty-function))

(defun (setf pretty-function) (fn op)
  (setf (get op '%operator-pretty-function) fn))

(defun print-function (op)
  (get op '%operator-print-function))

(defun (setf print-function) (fn op)
  (setf (get op '%operator-print-function) fn))

(defun macro-expander (op)
  (get op '%operator-macro-expander))

(defun (setf macro-expander) (fn op)
  (setf (get op '%operator-macro-expander) fn))




;;; expression printing dispatch

(defparameter *output-stream* *standard-output*)

(defun print-expression (x &optional (stream *output-stream*))
  (declare (special *symbol-key* *function-call-printer*))
  (let ((*print-length* nil)
        (*print-level* nil))
    (cond ((primitivep x)
           (write (if (symbolp x) 
                      (funcall *symbol-key* x)
                      x)
                  :escape (stringp x)
                  :stream stream))
          ((expressionp x)
           (setq x (macroexpand-expression x))
           (destructuring-bind (operator . arguments) x
             (cond ((and *print-pretty*
                         (pretty-function operator))
                    (funcall (pretty-function operator) operator arguments stream))
                   ((print-function (operator x))
                    (funcall (print-function operator) operator arguments stream))
                   (t (funcall *function-call-printer* operator arguments stream)))))
          ((error "print-expression can't print ~s" x)))))




;;; symbol names

(defparameter *symbol-key* 'dash-to-underscore)

(defun dash-to-underscore (x)
  (substitute #\_ #\- (symbol-name x)))




;;; functions

(defparameter *function-call-printer* 'print-function-call)

(defun print-function-call (op args stream)
  (print-expression op stream)
  (let ((length=1 (= (length args) 1)))
    (cond (*print-pretty*
           (write-char #\space stream)
           (pprint-logical-block (stream nil :prefix "(" :suffix ")")
             (when args
               (print-expression (car args) stream)
               (dolist (arg (cdr args))
                 (write-char #\, stream)
                 (write-char #\space stream)
                 (pprint-newline :linear stream)
                 (print-expression arg stream)))))
          (t
           (write-char #\( stream)
           (when args
             (print-expression (car args) stream)
             (dolist (arg (cdr args))
               (write-char #\, stream)
               (print-expression arg stream)))
           (write-char #\) stream)))))

#+notnow
(defun print-function-call (op args stream)
  (print-expression op stream)
  (let ((length=1 (= (length args) 1)))
    (cond (*print-pretty*
           (write-char #\space stream)
           (pprint-logical-block (stream nil :prefix (if length=1 "" "(") :suffix (if length=1 "" ")"))
             (when args
               (print-expression (car args) stream)
               (dolist (arg (cdr args))
                 (write-char #\, stream)
                 (write-char #\space stream)
                 (pprint-newline :linear stream)
                 (print-expression arg stream)))))
          (t
           (unless (length=1)
             (write-char #\( stream))
           (when args
             (print-expression (car args) stream)
             (dolist (arg (cdr args))
               (write-char #\, stream)
               (print-expression arg stream)))
           (unless (length=1)
             (write-char #\) stream))))))



;; TODO: add a statementp check to decide whether to print ; or not
(defun print-block (statements stream)
  (cond (*print-pretty*
         (when statements
           (pprint-newline :mandatory stream))
         (write-char #\{ stream)
         (pprint-logical-block (stream statements)
           (when statements
             (pprint-indent :block 1 stream))
           (loop
            (pprint-exit-if-list-exhausted)
            (pprint-newline :mandatory stream)
            (let ((form (macroexpand-expression (pprint-pop))))
              (print-expression form stream)
              (unless (and (expressionp form)
                           (eq (operator-type (operator form)) :special-statement))
                (write-char #\; stream)))))
         (when statements
           (pprint-newline :mandatory stream))
         (write-char #\} stream))
        (t
         (write-char #\{ stream)
         (loop for form in statements
               do
               (setq form (macroexpand-expression form))
               (print-expression form stream)
               (unless (and (expressionp form)
                            (eq (operator-type (operator form)) :special-statement))
                 (write-char #\; stream)))
         (write-char #\} stream))))




;;; infix operators

(defun pprint-infix (op args stream)
  (assert (>= (length args) 2))
  (pprint-logical-block (stream nil :prefix "(" :suffix ")")
    (print-expression (car args) stream)
    (dolist (arg (cdr args))
      (write-char #\space stream)
      (pprint-newline :linear stream)
      (write-string (string op) stream)
      (write-char #\space stream)
      (print-expression arg stream))))

(defun print-infix (op args stream)
  (write-char #\( stream)
  (print-expression (car args) stream)
  (dolist (arg (cdr args))
    (write-string (string op) stream)
    (print-expression arg stream))
  (write-char #\) stream))

(defun def-infix-operator (op)
  (setf (operator-type op) :infix-operator)
  (setf (pretty-function op) #'pprint-infix)
  (setf (print-function op) #'print-infix))




;;; special forms

(defmacro def-special-form-statement (op)
  `(setf (operator-type ',op) :special-statement))

(defmacro def-special-form-expression (op)
  `(setf (operator-type ',op) :special-expression))

(defmacro def-special-form-pretty (op-sym (args-sym stream-sym) &body body)
  `(setf (pretty-function ',op-sym)
    #'(lambda (,op-sym ,args-sym ,stream-sym)
        (declare (ignorable ,op-sym))
        ,@body)))

(defmacro def-special-form-print (op-sym (args-sym stream-sym) &body body)
  `(setf (print-function ',op-sym)
    #'(lambda (,op-sym ,args-sym ,stream-sym)
        (declare (ignorable ,op-sym))
        ,@body)))




;;; macro-forms

(defmacro def-macro-form (op-sym (args-sym) &body body)
  `(progn
    (setf (operator-type ',op-sym) :macro-operator)
    (setf (macro-expander ',op-sym)
     #'(lambda (,args-sym)
         (declare (ignorable ,args-sym))
         ,@body))))

(defparameter *next-genvar-id* 0)

(defun genvar (&optional (string ""))
  "Use genvar to create temporary variable names when defining extension macros.
Genvar simply uses a prefix and a counter and concatenates them with STRING.
Counter starts over with every js-compile-file so genvars are not guaranteed
to be unique across files."
  (intern (format nil "gsym_~A~A" string (incf *next-genvar-id*))))

(defun macroexpand-expression (form)
  (cond ((and (expressionp form)
              (eq (operator-type (operator form)) :macro-operator))
         (funcall (macro-expander (operator form)) (arguments form)))
        (t
         form)))




;;; helpers

(defun mkstr (&rest args)
  (with-output-to-string (s)
    (dolist (a args) (princ a s))))

(defun symb (&rest args)
  (values (intern (apply #'mkstr args))))



;;; expressions.lisp ends here