;;; -*- mode: lisp; syntax: common-lisp; base: 10; package: JSGEN -*-
;;
;; extensions.lisp -- Extensions to s-expression based JavaScript
;;
;; Copyright (c) 2006, Jason Kantz <jason@kantz.com>
;; All rights reserved.
;;
;;; Code:

(in-package :jsgen)


;;; control

(def-special-form-statement |begin|)

(def-special-form-print |begin| (args stream)
  (loop for e in args
        do
        (setq e (macroexpand-expression e))
        (print-expression e stream)
        (unless (eq (operator-type (operator e)) :special-statement)
          (write-char #\; stream))))

(def-special-form-pretty |begin| (args stream)
  (pprint-logical-block (stream nil)
    (let ((first-expression (macroexpand-expression (car args))))
      ;; (pprint-newline :mandatory stream)
      (print-expression first-expression stream)
      (unless (eq (operator-type (operator first-expression)) :special-statement)
          (write-char #\; stream))
      (loop for e in (cdr args)
            do
            (setq e (macroexpand-expression e))
            (pprint-newline :mandatory stream)
            (print-expression e stream)
            (unless (eq (operator-type (operator e)) :special-statement)
              (write-char #\; stream)))
      (pprint-indent :block 0 stream))))

(def-macro-form |or| (args)
  `(\|\| ,@args))

(def-macro-form |and| (args)
  `(\&\& ,@args))

(def-macro-form |not| (args)
  `(|!| ,@args))

(def-macro-form |when| (args)
  (destructuring-bind (test . body) args
      `(|cond| (,test ,@body))))




;;; strings

(def-special-form-expression |esc_str|)

(def-special-form-print |esc_str| (args stream)
  (cond ((and (null (cdr args))
              (stringp (car args)))
         (print-expression (format nil "\"~a\"" (car args)) stream))
        (t (print-expression `(+ "\"" ,@args "\"") stream))))




;;; iteration

(def-macro-form |dotimes| (args)
  (destructuring-bind ((n count) . body) args
    `(|for| (|var| ,n 0) (< ,n ,count) (|++| ,n)
      ,@body)))

(def-macro-form |doarray| (args)
  (let ((i (genvar "i")))
    (destructuring-bind ((var array) . body) args
      `(|for| (|var| ,i 0) (< ,i (|dot| ,array |length|)) (|++| ,i)
        (|var| ,var (|aref| ,array ,i))
        ,@body))))




;;; html generation

(defun make-open-tag (tag &key (singleton nil))
  (if (atom tag)
      (format nil "<~a~{~a~}>" tag (and singleton '(#\/)))
      (destructuring-bind (tag . attributes) tag
        (if (null attributes)
            (format nil "<~a~{~a~}>" tag (and singleton '(#\/)))
            `(+
              ,(format nil "<~a" tag)
              ,@(loop for (att val) on attributes by #'cddr
                      append
                      (cond ((stringp val)
                             (list (format nil " ~a=\"~a\"" att val)))
                            (t
                             (list (format nil " ~a=" att)
                                   `(|esc_str| ,val)))))
              ,(if singleton "/>" ">"))))))

(defun make-close-tag (tag)
  (format nil "</~a>" (if (consp tag)
                          (car tag)
                          tag)))

(defun html-form-p (x)
  (or (and (consp x)
           (consp (car x))
           (keywordp (caar x)))
      (and (consp x)
           (keywordp (car x)))))

(def-macro-form |html| (args)
  (destructuring-bind (var lhtml) args
      (cond
        ;; tag with attributes
        ((html-form-p lhtml)
         (destructuring-bind (tag . body) lhtml
           `(|begin|
             (+= ,var ,(make-open-tag tag :singleton (not body)))
             ,@(loop for x in body
                     collect
                     (cond ((html-form-p x)
                            `(|html| ,var ,x))
                           #+holdon
                           ((stringp x)
                            `(+= ,var ,x))
                           (t
                            `(+= ,var ,x))))
             ,@(when body
                  `((+= ,var ,(make-close-tag tag)))))))
        ((keywordp lhtml)
         (make-open-tag lhtml :singleton t))
        (t lhtml))))




;;; backquote unquote unquote-splicing

(defun insert-bq-nesting (bq-form &optional (level 0))
  (cond ((and (consp bq-form)
              (eq (car bq-form) '|bq|))
         `(|%bq| ,(insert-bq-nesting (cadr bq-form) (1+ level)) ,level))
        ((and (consp bq-form)
              (eq (car bq-form) '|uq|))
         `(|%uq| ,(insert-bq-nesting (cadr bq-form) level) ,(1- level)))
        ((consp bq-form)
         (mapcar #'(lambda (x) (insert-bq-nesting x level)) bq-form))
        ((and (stringp bq-form) (> (1- level) 0))
         (with-output-to-string (string)
           (print-esc-quote string (1- level))
           (format string "~a" bq-form)
           (print-esc-quote string (1- level))))
        (t
         bq-form)))

(def-macro-form |bq| (args)
  (insert-bq-nesting `(|bq| ,(car args))))

(def-macro-form |uq| (args)
  (error "uq appears outside of bq")
  nil)

(def-special-form-expression |%bq|)

(defun print-esc-quote (stream level)
  (cond ((= level 0)
         (write-char #\" stream))
        (t
         (dotimes (i (* (1- level) 2 ))
           (write-char #\\ stream))
         (write-string "\\\"" stream))))

(def-special-form-print |%bq| (args stream)
  (destructuring-bind (form level) args
    (unless (zerop level)
      (write-string "eval " stream))
    (pprint-logical-block (stream nil :prefix "(" :suffix ")")
      (print-esc-quote stream level)
      (print-expression form stream)
      (print-esc-quote stream level))))

(def-special-form-print |%uq| (args stream)
  (destructuring-bind (form level) args
    (print-esc-quote stream level)
    (write-string " + " stream)
    (pprint-logical-block (stream nil :prefix "(" :suffix ")")
      (if (stringp form)
          (write-string form stream)
          (print-expression form stream)))
    (write-string " + " stream)
    (print-esc-quote stream level)))




;;; aliases

(def-macro-form |set| (args)
  (destructuring-bind (var val) args
    `(|=| ,var ,val)))

(def-macro-form |get| (args)
  (destructuring-bind (var val) args
    `(|.| ,var ,val)))




;;; multi functions

(def-macro-form |mfunction| (args)
  (flet ((array-name (fn-name)
           (symb (string fn-name) "__function_list"))
         (found-name (fn-name)
           (symb (string fn-name) "__function_found")))
    (let ((result (genvar "result"))
          (fn (genvar "fn")))
      (destructuring-bind (fn-name fn-args . body) args
          (cond ((null body)
                 `(|begin|
                   (|var| ,(found-name fn-name) |undefined|)
                   (|var| ,(array-name fn-name) (|array|))
                   (|function| ,fn-name ,fn-args
                    (|cond| ((|==| ,(found-name fn-name) |undefined|)
                             (|var| ,result)
                             (|doarray| (,fn ,(array-name fn-name))
                                        (|=| ,result ,(if fn-args
                                                          `(,fn ,@fn-args)
                                                          `(,fn)))
                                        (|cond| ((|!=| ,result |undefined|)
                                                 (|=| ,(found-name fn-name) ,fn)
                                                 (|return| ,result)))))
                     (|else|
                      (|return| ,(if fn-args
                                     `(,(found-name fn-name) ,fn-args)
                                     `(,(found-name fn-name))))))
                    (|return| |undefined|))))
                (t
                 `(,(symb (array-name fn-name) ".push") (|function| ,fn-args
                                                     (|try| ,@body)
                                                     (|catch| |e|
                                                      (|return| |undefined|))))))))))



;;; extensions.lisp ends here