;;; -*- mode: lisp; syntax: common-lisp; base: 10; package: JSGEN -*-
;;
;; javascript.lisp -- JavaScript Language (ECMA 262) mapped to s-expressions.
;;
;; Copyright (c) 2006, Jason Kantz <jason@kantz.com>
;; All rights reserved.
;;
;;; Code:

(in-package :jsgen)


;;; translation functions

(defun js (stream code)
  "Translate CODE to JavaScript, outputing JavaScript to STREAM."
  (print-expression code stream))


(defun js-string (form)
  "Translate CODE to JavaScript, returning JavaScript as a string."
  (with-output-to-string (*output-stream*)
    (print-expression form *output-stream*)))

(defun js-compile-file (sjs-file &key (force nil))
  "Translate sjs file to JavaScript, creating a new file with extension js."
  (let ((js-file (make-pathname :type "js" :defaults sjs-file))
        (rtc (readtable-case *readtable*))
        (*package* (find-package :jsgen))
        (*next-genvar-id* 0))
    (unwind-protect 
         (when (or force
                   (not (probe-file js-file))
                   (> (file-write-date sjs-file)
                      (file-write-date js-file)))
           (setf (readtable-case *readtable*) :preserve)
           (with-open-file (sjs sjs-file :direction :input)
             (with-open-file (js js-file :direction :output :if-exists :supersede)
               (format js "// Auto Generated File.  It would be unwise to edit this file.~%~%")
               (loop with eof = (gensym)
                     as form = (read sjs nil eof)
                     until (eq form eof)
                     do
                     (print-expression form js)
                     (terpri js)
                     (terpri js)))))
      (setf (readtable-case *readtable*) rtc))))




;;; infix ops

(dolist (op '(* / % + << >> >>>
              < > <= >= |in| |instanceof|
              == != === !==
              = *= /= %= += -= <<= >>= >>>= \&= \|= ^=
              & ^ \| \&\& \|\|
              \,))
  (def-infix-operator op))




;;; other ops

(def-special-form-expression |.|)

(def-special-form-print |.| (args stream)
  (print-expression (car args) stream)
  (dolist (arg (cdr args))
    (unless (and (symbolp arg) (char= (schar (symbol-name arg) 0) #\[))
      (write-string "." stream))
    (print-expression arg stream)))


(def-special-form-expression |++|)

(def-special-form-print |++| (args stream)
  (format stream "++~a" (car args)))


(def-special-form-expression |_++|)

(def-special-form-print |_++| (args stream)
  (format stream "~a++" (car args)))


(def-special-form-expression |--|)

(def-special-form-print |--| (args stream)
  (format stream "--~a" (car args)))


(def-special-form-expression |_--|)

(def-special-form-print |_--| (args stream)
  (format stream "~a--" (car args)))


(def-special-form-expression |-|)

(def-special-form-print |-| (args stream)
  (if (null (cdr args))
      (progn
        (write-char #\- stream)
        (print-expression (car args) stream))
      (print-infix '|-| args stream)))


(def-special-form-pretty |-| (args stream)
  (if (null (cdr args))
      (progn
        (write-char #\- stream)
        (print-expression (car args) stream))
      (pprint-infix '|-| args stream)))


(def-special-form-expression |?|)

(def-special-form-print |?| (args stream)
  (write-char #\( stream)
  (write-char #\( stream)
  (print-expression (pop args) stream)
  (write-char #\) stream)
  (write-string " ? " stream)
  (write-char #\( stream)
  (print-expression (pop args) stream)
  (write-char #\) stream)
  (write-string " : " stream)
  (write-char #\( stream)
  (print-expression (pop args) stream)
  (write-char #\) stream)
  (write-char #\) stream))


(def-special-form-expression |!|)

(def-special-form-print |!| (args stream)
  (write-char #\! stream)
  (print-expression (car args) stream))


(def-special-form-expression |~|)

(def-special-form-print |~| (args stream)
  (write-char #\~ stream )
  (print-expression (car args) stream))




;;; aref

(def-special-form-expression |aref|)

;; (set favorite (aref coffees 1))

(def-special-form-print |aref| (args stream)
  (print-expression (pop args) stream)
  (loop for subscript in args
        do
        (write-char #\[ stream)
        (print-expression (pop args) stream)
        (write-char #\] stream)))




;;; array

(def-special-form-expression |array|)

;; (set coffees (array "French Roast" "Columbian" "Kona"))

(def-special-form-print |array| (args stream)
  (write-char #\[ stream)
  (when args
    (print-expression (pop args) stream)
    (loop while args do
          (write-string "," stream)
          (print-expression (pop args) stream)))
  (write-char #\] stream))

(def-special-form-pretty |array| (args stream)
  (pprint-logical-block (stream args :prefix "[" :suffix "]")
    (pprint-exit-if-list-exhausted)
    (print-expression (pprint-pop) stream)
    (loop
     (pprint-exit-if-list-exhausted)
     (write-string "," stream)
     (pprint-newline :linear stream)
     (print-expression (pprint-pop) stream))))




;;; break

(def-special-form-statement |break|)

(def-special-form-print |break| (args stream)
  (write-string "break" stream)
  (when args
    (write-char #\Space stream)
    (print-expression (car args) stream))
  (write-char #\; stream))




;;; cond

(def-special-form-statement |cond|) 

(def-special-form-print |cond| (args stream)
  (let* ((case (car args))
         (test (car case)))
    (write-string "if" stream)
    (unless (consp test)
      (write-char #\( stream))
    (print-expression test stream)
    (unless (consp test)
      (write-char #\) stream))
    (print-block (cdr case) stream))
  (loop for case in (cdr args)
        as test = (car case)
        do
        (cond ((eq test '|else|)
               (write-string "else" stream))
              (t (write-string "else if" stream)
                 (unless (consp test)
                   (write-char #\( stream))
                 (print-expression test stream)
                 (unless (consp test)
                   (write-char #\) stream))))
        (print-block (cdr case) stream)))

(def-special-form-pretty |cond| (args stream)
  (pprint-logical-block (stream nil)
    (let* ((case (car args))
           (test (car case)))
      (write-string "if " stream)
      (if (consp test)
          (pprint-logical-block (stream nil)
            (print-expression test stream))
          (pprint-logical-block (stream nil :prefix "(" :suffix ")")
            (print-expression test stream)))
      (print-block (cdr case) stream))
    (when (cdr args)
      (loop for case in (cdr args)
            as test = (car case)
            do
            (pprint-newline :mandatory stream)
            (cond ((eq test '|else|)
                   (write-string "else" stream))
                  (t (write-string "else if " stream)
                     (if (consp test)
                         (pprint-logical-block (stream nil)
                           (print-expression test stream))
                         (pprint-logical-block (stream nil :prefix "(" :suffix ")")
                           (print-expression test stream)))))
            (print-block (cdr case) stream)))))




;;; continue

(def-special-form-statement |continue|)

(def-special-form-print |continue| (args stream)
  (write-string "continue" stream)
  (when args
    (write-char #\Space stream)
    (print-expression (car args) stream))
  (write-char #\; stream))




;;; do

(def-special-form-statement |do|)

(def-special-form-print |do| (args stream)
  (write-string "do" stream)
  (print-block (subseq args 0 (1- (length args))) stream)
  (write-string "while" stream)
  (write-char #\( stream)
  (print-expression (second (car (last args))) stream)
  (write-char #\) stream)
  (write-char #\; stream))

(def-special-form-pretty |do| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "do " stream)
    (print-block (subseq args 0 (1- (length args))) stream)
    (write-string " while " stream)
    (pprint-logical-block (stream nil :prefix "(" :suffix ")")
      (print-expression (second (car (last args))) stream))
    (write-char #\; stream)
    (pprint-newline :mandatory stream)))




;;; for

(def-special-form-statement |for|)

(def-special-form-print |for| (args stream)
  (destructuring-bind (init-statement test update . body) args
    (write-string "for(" stream)
    (write-char #\( stream)
    (cond ((eq test '|in|)
           (print-expression init-statement stream)
           (write-string " in " stream)
           (print-expression update stream))
          (t
           (print-expression init-statement stream)
           (write-string " " stream)
           (print-expression test stream)
           (write-string "; " stream)
           (print-expression update stream)))
    (write-char #\) stream)
    (print-block body stream)))

(def-special-form-pretty |for| (args stream)
  (destructuring-bind (init-statement test update . body) args
    (pprint-logical-block (stream nil)
      (write-string "for " stream)
      (pprint-logical-block (stream nil :prefix "(" :suffix ")")
        (cond ((eq test '|in|)
               (print-expression init-statement stream)
               (write-string " in " stream)
               (print-expression update stream))
              (t
               (print-expression init-statement stream)
               (write-string " " stream)
               (pprint-newline :linear stream)
               (print-expression test stream)
               (write-string "; " stream)
               (pprint-newline :linear stream)
               (print-expression update stream))))
      (print-block body stream))))




;;; function

(def-special-form-expression |function|)

(def-special-form-print |function| (args stream)
  (write-string "function " stream)
  (when (and (symbolp (car args))
             (not (null (car args))))
    (print-expression (pop args) stream))
  (write-char #\( stream)
  (loop with fn-args = (pop args)
        while fn-args
        do
        (print-expression (pop fn-args) stream)
        (when fn-args
          (write-char #\, stream)))
  (write-char #\) stream)
  (print-block args stream))

(def-special-form-pretty |function| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "function " stream)
    (when (and (symbolp (car args))
               (not (null (car args))))
      (print-expression (pop args) stream))
    (pprint-logical-block (stream (pop args) :prefix "(" :suffix ")")
      (pprint-exit-if-list-exhausted)
      (print-expression (pprint-pop) stream)
      (loop
       (pprint-exit-if-list-exhausted)
       (write-char #\, stream)
       (write-char #\space stream)
       (pprint-newline :linear stream)
       (print-expression (pprint-pop) stream)))
    (print-block args stream)))




;;; label

(def-special-form-statement |label|)

(def-special-form-print |label| (args stream)
  (print-expression (car args) stream)
  (write-string ": " stream)
  (print-expression (cadr args) stream)
  (unless (eq (operator-type (operator (cadr args))) :special-statement)
    (write-char #\; stream)))

(def-special-form-pretty |label| (args stream)
  (pprint-logical-block (stream nil)
    (print-expression (car args) stream)
    (write-string ": " stream)
    (pprint-newline :mandatory stream)
    (print-expression (cadr args) stream)
    (unless (eq (operator-type (operator (cadr args))) :special-statement)
      (write-char #\; stream))))




;;; new

(def-special-form-expression |new|)

;; (js #%(dot (new Date) (toString)))

(def-special-form-print |new| (args stream)
  (write-string "new " stream)
  (print-expression args stream))




;;; object

(def-special-form-expression |object|)

;; Example: (object :size 10 :type "shoe")

(def-special-form-print |object| (args stream)
  (write-char #\{ stream)
  (when args
    (print-expression (pop args) stream)
    (write-string ":" stream)
    (print-expression (pop args) stream))
  (loop while args do
        (write-string "," stream)
        (print-expression (pop args) stream)
        (write-string ":" stream)
        (print-expression (pop args) stream))
  (write-char #\} stream))

(def-special-form-pretty |object| (args stream)
  (pprint-logical-block (stream args :prefix "{" :suffix "}")
    (pprint-exit-if-list-exhausted)
    (print-expression (pprint-pop) stream)
    (write-string ":" stream)
    (print-expression (pprint-pop) stream)
    (loop
     (pprint-exit-if-list-exhausted)
     (write-string "," stream)
     (pprint-newline :linear stream)
     (print-expression (pprint-pop) stream)
     (write-string ":" stream)
     (print-expression (pprint-pop) stream))))




;;; return

(def-special-form-statement |return|)

(def-special-form-print |return| (args stream)
  (write-string (format nil "return") stream)
  (when args
    (write-char #\Space stream)
    (print-expression (car args) stream))
  (write-char #\; stream))




;;; switch

(def-special-form-statement |switch|)

;; (switch string-or-number-form (case string-or-number . body)* (default . body)?)

(def-special-form-print |switch| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "switch(" stream)
    (print-expression (pop args) stream)
    (write-char #\) stream)
    (print-block args stream)))

(def-special-form-pretty |switch| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "switch " stream)
    (pprint-logical-block (stream nil :prefix "(" :suffix ")")
      (print-expression (pop args) stream))
    (print-block args stream)))

(def-special-form-statement |case|) 

(def-special-form-print |case| (args stream)
  (write-string "case " stream)
  (print-expression (pop args) stream)
  (write-string ":" stream)
  (loop while args
        do
        (let ((statement (pop args)))
          (print-expression statement stream)
          (unless (and (consp statement)
                       (eq (operator-type (operator statement))
                           :special-statement))
            (write-char #\; stream)))))

(def-special-form-pretty |case| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "case " stream)
    (print-expression (pop args) stream)
    (write-string ": " stream)
    (pprint-indent :block 2 stream)
    (pprint-newline :mandatory stream)
    (loop while args
          do
          (let ((statement (pop args)))
            (print-expression statement stream)
            (unless (and (consp statement)
                         (eq (operator-type (operator statement))
                             :special-statement))
              (write-char #\; stream))
            (pprint-newline :mandatory stream)
            (unless args
              (pprint-indent :block 0 stream))))))

(def-special-form-statement |default|) 

(def-special-form-print |default| (args stream)
  (write-string "default:" stream)
  (loop while args
        do
        (let ((statement (pop args)))
          (print-expression statement stream)
          (unless (and (consp statement)
                       (eq (operator-type (operator statement))
                           :special-statement))
            (write-char #\; stream)))))

(def-special-form-pretty |default| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "default: " stream)
    (pprint-indent :block 2 stream)
    (pprint-newline :mandatory stream)
    (loop while args
          do
          (let ((statement (pop args)))
            (print-expression statement stream)
            (write-char #\; stream)
            (unless args
              (pprint-indent :block 0 stream))))))




;;; try

(def-special-form-statement |try|)

(def-special-form-print |try| (args stream)
  (write-string "try " stream)
  (print-block args stream))

(def-special-form-pretty |try| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "try " stream)
    (print-block args stream)))

(def-special-form-statement |catch|)

(def-special-form-print |catch| (args stream)
  (write-string " catch(" stream)
  (print-expression (car args) stream)
  (write-char #\) stream)
  (print-block (cdr args) stream))

(def-special-form-pretty |catch| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "catch " stream)
    (pprint-logical-block (stream nil :prefix "(" :suffix ")")
      (print-expression (car args) stream))
    (print-block (cdr args) stream)))

(def-special-form-statement |finally|)

(def-special-form-print |catch| (args stream)
  (write-string " finally" stream)
  (print-block (cdr args) stream))

(def-special-form-pretty |finally| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "finally " stream)
    (print-block args stream)))




;;; var

(def-special-form-statement |var|)

;; (var x (/ Math.PI 2) y (z (/ Math.PI 4)))
;; (var z)

(def-special-form-print |var| (args stream)
  (write-string "var " stream)
  (loop for (var val next) on args by #'cddr
        do
        (cond (val
               (print-expression var stream)
               (write-string " = " stream)
               (print-expression val stream))
              (t
               (print-expression var stream)))
        (when next
          (write-string ", " stream)))
  (write-char #\; stream))




;;; while

(def-special-form-statement |while|)

(def-special-form-print |while| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "while" stream)
    (write-char #\( stream)
    (print-expression (pop args) stream)
    (write-char #\) stream)
    (print-block args stream)))

(def-special-form-pretty |while| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "while " stream)
    (pprint-logical-block (stream nil :prefix "(" :suffix ")")
      (print-expression (pop args) stream))
    (print-block args stream)))




;;; with

(def-special-form-statement |with|)

;; (with Math
;;   (= a (* PI r r))
;;   (= x (* r (cos PI)))
;;   (= y (* r (sin (/ PI 2)))))

(def-special-form-print |with| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "with" stream)
    (write-char #\( stream)
    (print-expression (pop args) stream)
    (write-char #\) stream)
    (print-block args stream)))

(def-special-form-pretty |with| (args stream)
  (pprint-logical-block (stream nil)
    (write-string "with " stream)
    (pprint-logical-block (stream nil :prefix "(" :suffix ")")
      (print-expression (pop args) stream))
    (print-block args stream)))


;;; javascript.lisp ends here